/*-
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] =
"@(#) Copyright (c) 1988, 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
#if 0
static char sccsid[] = "@(#)kdump.c	8.1 (Berkeley) 6/6/93";
#endif
static const char rcsid[] =
  "$FreeBSD: src/usr.bin/kdump/kdump.c,v 1.16 1999/08/28 01:02:21 peter Exp $";
#endif /* not lint */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#define _KERNEL
extern int errno;
#include <sys/errno.h>
#undef _KERNEL
#include <sys/param.h>
#include <sys/ioctl.h>
#define _KERNEL
#include <sys/time.h>
#undef _KERNEL
#include <sys/uio.h>
#include <sys/ktrace.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/queue.h>
#include <sys/sysent.h>
#define _KERNEL
#include <sys/signalvar.h>
#undef _KERNEL
#include <sys/un.h>

#ifdef IPX
#include <sys/types.h>
#include <netipx/ipx.h>
#endif
#ifdef NETATALK
#include <netatalk/at.h>
#endif
#include <netinet/in.h>
#include <arpa/inet.h>

#include <ctype.h>
#include <err.h>
#include <grp.h>
#include <inttypes.h>
#include <locale.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <vis.h>

#include "kdump.h"
#include "ktrace.h"
#include "kdump_subr.h"

u_int abidump(struct ktr_header *);
int fetchprocinfo(struct ktr_header *, u_int *);
int fread_tail(void *, int, int);
void dumpheader(struct ktr_header *);
void ktrsyscall(struct ktr_syscall *);
void ktrsysret(struct ktr_sysret *);
void ktrnamei(char *, int);
void hexdump(char *, int, int);
void visdump(char *, int, int);
void ktrpsig(struct ktr_psig *);
void ktrgenio(struct ktr_genio *, int);
void ktrcsw(struct ktr_csw *);
void ktrstruct(char *buf, size_t buflen);
void ktruser(int, unsigned char *);
void sockfamilyname (int arg);
void usage(void);

void linux_ktrsyscall(struct ktr_syscall *);
void linux_ktrsysret(struct ktr_sysret *);
void linux_ktrstruct(char *buf, size_t buflen);

int timestamp, decimal, fancy = 1, suppressdata, tail, threads, maxdata,
    resolv = 0, abiflag = 0;
char *tracefile = DEF_TRACEFILE;

struct proc_info
{
	TAILQ_ENTRY(proc_info)	info;
	u_int			sv_flags;
	pid_t			pid;
};

TAILQ_HEAD(trace_procs, proc_info) trace_procs;

int
main(int argc, char *argv[])
{
	int ch, ktrlen, size;
	register void *m;
	int trpoints = ALL_POINTS;
	struct ktr_header ktr_header;
	pid_t pid = 0;
	int drop_logged;
	u_int sv_flags;

	setlocale(LC_CTYPE, "");

	while ((ch = getopt(argc,argv,"f:dElm:np:AHRrsTt:")) != -1)
		switch((char)ch) {
                case 'A':
			abiflag = 1;
			break;
		case 'f':
			tracefile = optarg;
			break;
		case 'd':
			decimal = 1;
			break;
		case 'l':
			tail = 1;
			break;
		case 'm':
			maxdata = atoi(optarg);
			break;
		case 'n':
			fancy = 0;
			break;
		case 'p':
			pid = atoi(optarg);
			break;
		case 'r':
			resolv = 1;
			break;
		case 's':
			suppressdata = 1;
			break;
		case 'E':
			timestamp = 3;	/* elapsed timestamp */
			break;
		case 'H':
			threads = 1;
			break;
		case 'R':
			timestamp = 2;	/* relative timestamp */
			break;
		case 'T':
			timestamp = 1;
			break;
		case 't':
			trpoints = getpoints(optarg);
			if (trpoints < 0)
				errx(1, "unknown trace point in %s", optarg);
			break;
		default:
			usage();
		}

	if (argc > optind)
		usage();

	m = (void *)malloc(size = 1025);
	if (m == NULL)
		errx(1, "%s", strerror(ENOMEM));
	if (!freopen(tracefile, "r", stdin))
		err(1, "%s", tracefile);
	TAILQ_INIT(&trace_procs);
	drop_logged = 0;
	while (fread_tail(&ktr_header, sizeof(struct ktr_header), 1)) {
		if (ktr_header.ktr_type & KTR_DROP) {
			ktr_header.ktr_type &= ~KTR_DROP;
			if (!drop_logged && threads) {
				printf("%6d %6ld %-8.*s Events dropped.\n",
				    ktr_header.ktr_pid, ktr_header.ktr_tid >
				    0 ? (u_long)ktr_header.ktr_tid : 0, MAXCOMLEN,
				    ktr_header.ktr_comm);
				drop_logged = 1;
			} else if (!drop_logged) {
				printf("%6d %-8.*s Events dropped.\n",
				    ktr_header.ktr_pid, MAXCOMLEN,
				    ktr_header.ktr_comm);
				drop_logged = 1;
			}
		}
		if (trpoints & (1<<ktr_header.ktr_type))
			if (pid == 0 || ktr_header.ktr_pid == pid)
				dumpheader(&ktr_header);
		if ((ktrlen = ktr_header.ktr_len) < 0)
			errx(1, "bogus length 0x%x", ktrlen);
		if (ktrlen > size) {
			m = (void *)realloc(m, ktrlen+1);
			if (m == NULL)
				errx(1, "%s", strerror(ENOMEM));
			size = ktrlen;
		}
		if (ktrlen && fread_tail(m, ktrlen, 1) == 0)
			errx(1, "data too short");
		if (fetchprocinfo(&ktr_header, (u_int *)m) != 0)
			continue;
		sv_flags = abidump(&ktr_header);	
		if (pid && ktr_header.ktr_pid != pid)
			continue;
		if ((trpoints & (1<<ktr_header.ktr_type)) == 0)
			continue;
		drop_logged = 0;
		switch (ktr_header.ktr_type) {
		case KTR_SYSCALL:
			if ((sv_flags & SV_ABI_MASK) == SV_ABI_LINUX)
				linux_ktrsyscall((struct ktr_syscall *)m);
			else
				ktrsyscall((struct ktr_syscall *)m);
			break;
		case KTR_SYSRET:
			if ((sv_flags & SV_ABI_MASK) == SV_ABI_LINUX)
				linux_ktrsysret((struct ktr_sysret *)m);
			else
				ktrsysret((struct ktr_sysret *)m);
			break;
		case KTR_NAMEI:
			ktrnamei(m, ktrlen);
			break;
		case KTR_GENIO:
			ktrgenio((struct ktr_genio *)m, ktrlen);
			break;
		case KTR_PSIG:
			ktrpsig((struct ktr_psig *)m);
			break;
		case KTR_CSW:
			ktrcsw((struct ktr_csw *)m);
			break;
		case KTR_USER:
			ktruser(ktrlen, m);
			break;
		case KTR_STRUCT:
			if ((sv_flags & SV_ABI_MASK) == SV_ABI_FREEBSD)
				ktrstruct(m, ktrlen);
			else if ((sv_flags & SV_ABI_MASK) == SV_ABI_LINUX)
				linux_ktrstruct(m, ktrlen);
			else
				ktrstruct(m, ktrlen);
			break;
		default:
			printf("\n");
			break;
		}
		if (tail)
			fflush(stdout);
	}
	return (0);
}

int
fread_tail(void *buf, int size, int num)
{
	int i;

	while ((i = fread(buf, size, num, stdin)) == 0 && tail) {
		(void)sleep(1);
		clearerr(stdin);
	}
	return (i);
}

int
fetchprocinfo(struct ktr_header *kth, u_int *flags)
{
	struct proc_info *pi;

	switch (kth->ktr_type) {
	case KTR_PROCCTOR:
		TAILQ_FOREACH(pi, &trace_procs, info) {
			if (pi->pid == kth->ktr_pid) {
				TAILQ_REMOVE(&trace_procs, pi, info);
				break;
			}
		}
		pi = malloc(sizeof(struct proc_info));
		if (pi == NULL)
			errx(1, "%s", strerror(ENOMEM));
		pi->sv_flags = *flags;
		pi->pid = kth->ktr_pid;
		TAILQ_INSERT_TAIL(&trace_procs, pi, info);
		return (1);

	case KTR_PROCDTOR:
		TAILQ_FOREACH(pi, &trace_procs, info) {
			if (pi->pid == kth->ktr_pid) {
				TAILQ_REMOVE(&trace_procs, pi, info);
				free(pi);
				break;
			}
		}
		return (1);
	}

	return (0);
}

u_int
abidump(struct ktr_header *kth)
{
	struct proc_info *pi;
	const char *abi;
	const char *arch;
	u_int flags = 0;

	TAILQ_FOREACH(pi, &trace_procs, info) {
		if (pi->pid == kth->ktr_pid) {
			flags = pi->sv_flags;
			break;
		}
	}

	if (abiflag == 0)
		return (flags);

	switch (flags & SV_ABI_MASK) {
	case SV_ABI_LINUX:
		abi = "L";
		break;
	case SV_ABI_FREEBSD:
		abi = "F";
		break;
	default:
		abi = "U";
		break;
	}

	if (flags != 0) {
		if (flags & SV_LP64)
			arch = "64";
		else
			arch = "32";
	} else
		arch = "00";

	printf("%s%s  ", abi, arch);

	return (flags);
}

void
dumpheader(struct ktr_header *kth)
{
	static char unknown[64];
	static struct timeval prevtime, temp;
	const char *type;

	switch (kth->ktr_type) {
	case KTR_SYSCALL:
		type = "CALL";
		break;
	case KTR_SYSRET:
		type = "RET ";
		break;
	case KTR_NAMEI:
		type = "NAMI";
		break;
	case KTR_GENIO:
		type = "GIO ";
		break;
	case KTR_PSIG:
		type = "PSIG";
		break;
	case KTR_CSW:
		type = "CSW ";
		break;
	case KTR_USER:
		type = "USER";
		break;
	case KTR_STRUCT:
		type = "STRU";
		break;
	case KTR_SYSCTL:
		type = "SCTL";
		break;
	case KTR_PROCCTOR:
		/* FALLTHROUGH */
	case KTR_PROCDTOR:
		return;
	default:
		sprintf(unknown, "UNKNOWN(%d)", kth->ktr_type);
		type = unknown;
	}

	/*
	 * The ktr_tid field was previously the ktr_buffer field, which held
	 * the kernel pointer value for the buffer associated with data
	 * following the record header.  It now holds a threadid, but only
	 * for trace files after the change.  Older trace files still contain
	 * kernel pointers.  Detect this and suppress the results by printing
	 * negative tid's as 0.
	 */
	if (threads)
		printf("%6jd %6jd %-8.*s ", (intmax_t)kth->ktr_pid,
		    kth->ktr_tid > 0 ? (intmax_t)kth->ktr_tid : 0,
		    MAXCOMLEN, kth->ktr_comm);
	else
		printf("%6jd %-8.*s ", (intmax_t)kth->ktr_pid, MAXCOMLEN,
		    kth->ktr_comm);
	if (timestamp) {
		if (timestamp == 3) {
			if (prevtime.tv_sec == 0)
				prevtime = kth->ktr_time;
			timevalsub(&kth->ktr_time, &prevtime);
		}
		if (timestamp == 2) {
			temp = kth->ktr_time;
			timevalsub(&kth->ktr_time, &prevtime);
			prevtime = temp;
		}
		printf("%jd.%06ld ", (intmax_t)kth->ktr_time.tv_sec,
		    kth->ktr_time.tv_usec);
	}
	printf("%s  ", type);
}

#include <sys/syscall.h>
#define KTRACE
#include <kern/syscalls.c>
#undef KTRACE
int nsyscalls = sizeof (syscallnames) / sizeof (syscallnames[0]);

void
ktrsyscall(struct ktr_syscall *ktr)
{
	int narg = ktr->ktr_narg;
	register_t *ip;

	if (ktr->ktr_code >= nsyscalls || ktr->ktr_code < 0)
		printf("[%d]", ktr->ktr_code);
	else
		printf("%s", syscallnames[ktr->ktr_code]);
	ip = &ktr->ktr_args[0];
	if (narg) {
		char c = '(';
		while (narg > 0)
			print_number(ip,narg,c);
		putchar(')');
	}
	putchar('\n');
}

void
ktrsysret(struct ktr_sysret *ktr)
{
	register_t ret = ktr->ktr_retval;
	int error = ktr->ktr_error;
	int code = ktr->ktr_code;

	if (code >= nsyscalls || code < 0)
		printf("[%d] ", code);
	else
		printf("%s ", syscallnames[code]);

	if (error == 0) {
		if (fancy) {
			printf("%ld", (long)ret);
			if (ret < 0 || ret > 9)
				printf("/%#lx", (long)ret);
		} else {
			if (decimal)
				printf("%ld", (long)ret);
			else
				printf("%#lx", (long)ret);
		}
	} else if (error == ERESTART)
		printf("RESTART");
	else if (error == EJUSTRETURN)
		printf("JUSTRETURN");
	else {
		printf("-1 errno %d", ktr->ktr_error);
		if (fancy)
			printf(" %s", strerror(ktr->ktr_error));
	}
	putchar('\n');
}

void
ktrnamei(char *cp, int len)
{
	printf("\"%.*s\"\n", len, cp);
}

void
hexdump(char *p, int len, int screenwidth)
{
	int n, i;
	int width;

	width = 0;
	do {
		width += 2;
		i = 13;			/* base offset */
		i += (width / 2) + 1;	/* spaces every second byte */
		i += (width * 2);	/* width of bytes */
		i += 3;			/* "  |" */
		i += width;		/* each byte */
		i += 1;			/* "|" */
	} while (i < screenwidth);
	width -= 2;

	for (n = 0; n < len; n += width) {
		for (i = n; i < n + width; i++) {
			if ((i % width) == 0) {	/* beginning of line */
				printf("       0x%04x", i);
			}
			if ((i % 2) == 0) {
				printf(" ");
			}
			if (i < len)
				printf("%02x", p[i] & 0xff);
			else
				printf("  ");
		}
		printf("  |");
		for (i = n; i < n + width; i++) {
			if (i >= len)
				break;
			if (p[i] >= ' ' && p[i] <= '~')
				printf("%c", p[i]);
			else
				printf(".");
		}
		printf("|\n");
	}
	if ((i % width) != 0)
		printf("\n");
}

void
visdump(char *dp, int datalen, int screenwidth)
{
	int col = 0;
	char *cp;
	int width;
	char visbuf[5];

	(void)printf("       \"");
	col = 8;
	for (;datalen > 0; datalen--, dp++) {
		vis(visbuf, *dp, VIS_CSTYLE, *(dp+1));
		cp = visbuf;
		/*
		 * Keep track of printables and
		 * space chars (like fold(1)).
		 */
		if (col == 0) {
			putchar('\t');
			col = 8;
		}
		switch(*cp) {
		case '\n':
			col = 0;
			putchar('\n');
			continue;
		case '\t':
			width = 8 - (col&07);
			break;
		default:
			width = strlen(cp);
		}
		if (col + width > (screenwidth-2)) {
			printf("\\\n\t");
			col = 8;
		}
		col += width;
		do {
			putchar(*cp++);
		} while (*cp);
	}
	if (col == 0)
		printf("       ");
	printf("\"\n");
}

void
ktrgenio(struct ktr_genio *ktr, int len)
{
	int datalen = len - sizeof (struct ktr_genio);
	char *dp = (char *)ktr + sizeof (struct ktr_genio);
	static int screenwidth = 0;
	int i, binary;

	if (screenwidth == 0) {
		struct winsize ws;

		if (fancy && ioctl(fileno(stderr), TIOCGWINSZ, &ws) != -1 &&
		    ws.ws_col > 8)
			screenwidth = ws.ws_col;
		else
			screenwidth = 80;
	}
	printf("fd %d %s %d byte%s\n", ktr->ktr_fd,
		ktr->ktr_rw == UIO_READ ? "read" : "wrote", datalen,
		datalen == 1 ? "" : "s");
	if (suppressdata)
		return;
	if (maxdata && datalen > maxdata)
		datalen = maxdata;

	for (i = 0, binary = 0; i < datalen && binary == 0; i++)  {
		if (dp[i] >= 32 && dp[i] < 127)
			continue;
		if (dp[i] == 10 || dp[i] == 13 || dp[i] == 0 || dp[i] == 9)
			continue;
		binary = 1;
	}
	if (binary)
		hexdump(dp, datalen, screenwidth);
	else
		visdump(dp, datalen, screenwidth);
}

const char *signames[] = {
	"NULL", "HUP", "INT", "QUIT", "ILL", "TRAP", "IOT",	/*  1 - 6  */
	"EMT", "FPE", "KILL", "BUS", "SEGV", "SYS",		/*  7 - 12 */
	"PIPE", "ALRM",  "TERM", "URG", "STOP", "TSTP",		/* 13 - 18 */
	"CONT", "CHLD", "TTIN", "TTOU", "IO", "XCPU",		/* 19 - 24 */
	"XFSZ", "VTALRM", "PROF", "WINCH", "29", "USR1",	/* 25 - 30 */
	"USR2", NULL,						/* 31 - 32 */
};

void
ktrpsig(struct ktr_psig *psig)
{
	if (psig->signo > 0 && psig->signo < NSIG)
		printf("SIG%s ", signames[psig->signo]);
	else
		printf("SIG %d ", psig->signo);
	if (psig->action == SIG_DFL)
		printf("SIG_DFL\n");
	else {
		printf("caught handler=0x%lx mask=0x%x code=0x%x\n",
		    (u_long)psig->action, psig->mask.__bits[0], psig->code);
	}
}

void
ktrcsw(struct ktr_csw *cs)
{
	printf("%s %s\n", cs->out ? "stop" : "resume",
	    cs->user ? "user" : "kernel");
}

void
ktruser(int len, unsigned char *p)
{
	printf("%d ", len);
	while (len--)
		if (decimal)
			printf(" %d", *p++);
		else
			printf(" %02x", *p++);
	printf("\n");
}

void
ktrsockaddr(struct sockaddr *sa)
{
/*
 TODO: Support additional address families
	#include <netnatm/natm.h>
	struct sockaddr_natm	*natm;
	#include <netsmb/netbios.h>
	struct sockaddr_nb	*nb;
*/
	char addr[64];

	/*
	 * note: ktrstruct() has already verified that sa points to a
	 * buffer at least sizeof(struct sockaddr) bytes long and exactly
	 * sa->sa_len bytes long.
	 */
	printf("struct sockaddr { ");
	sockfamilyname(sa->sa_family);
	printf(", ");

#define check_sockaddr_len(n)					\
	if (sa_##n->s##n##_len < sizeof(struct sockaddr_##n)) {	\
		printf("invalid");				\
		break;						\
	}

	switch(sa->sa_family) {
	case AF_INET: {
		struct sockaddr_in	*sa_in;

		sa_in = (struct sockaddr_in *)sa;
		check_sockaddr_len(in);
		inet_ntop(AF_INET, &sa_in->sin_addr, addr, sizeof addr);
		printf("%s:%u", addr, ntohs(sa_in->sin_port));
		break;
	}
#ifdef NETATALK
	case AF_APPLETALK: {
		struct sockaddr_at	*sa_at;
		struct netrange		*nr;

		sa_at = (struct sockaddr_at *)sa;
		check_sockaddr_len(at);
		nr = &sa_at->sat_range.r_netrange;
		printf("%d.%d, %d-%d, %d", ntohs(sa_at->sat_addr.s_net),
			sa_at->sat_addr.s_node, ntohs(nr->nr_firstnet),
			ntohs(nr->nr_lastnet), nr->nr_phase);
		break;
	}
#endif
	case AF_INET6: {
		struct sockaddr_in6	*sa_in6;

		sa_in6 = (struct sockaddr_in6 *)sa;
		check_sockaddr_len(in6);
		inet_ntop(AF_INET6, &sa_in6->sin6_addr, addr, sizeof addr);
		printf("[%s]:%u", addr, htons(sa_in6->sin6_port));
		break;
	}
#ifdef IPX
	case AF_IPX: {
		struct sockaddr_ipx	*sa_ipx;

		sa_ipx = (struct sockaddr_ipx *)sa;
		check_sockaddr_len(ipx);
		/* XXX wish we had ipx_ntop */
		printf("%s", ipx_ntoa(sa_ipx->sipx_addr));
		break;
	}
#endif
	case AF_UNIX: {
		struct sockaddr_un *sa_un;

		sa_un = (struct sockaddr_un *)sa;
		check_sockaddr_len(un);
		printf("%.*s", (int)sizeof(sa_un->sun_path), sa_un->sun_path);
		break;
	}
	default:
		printf("unknown address family");
	}
	printf(" }\n");
}

void
ktrstat(struct stat *statp)
{
	char mode[12], timestr[PATH_MAX + 4];
	struct passwd *pwd;
	struct group  *grp;
	struct tm *tm;

	/*
	 * note: ktrstruct() has already verified that statp points to a
	 * buffer exactly sizeof(struct stat) bytes long.
	 */
	printf("struct stat {");
	strmode(statp->st_mode, mode);
	printf("dev=%ju, ino=%ju, mode=%s, nlink=%ju, ",
		(uintmax_t)statp->st_dev, (uintmax_t)statp->st_ino, mode,
		(uintmax_t)statp->st_nlink);
	if (resolv == 0 || (pwd = getpwuid(statp->st_uid)) == NULL)
		printf("uid=%ju, ", (uintmax_t)statp->st_uid);
	else
		printf("uid=\"%s\", ", pwd->pw_name);
	if (resolv == 0 || (grp = getgrgid(statp->st_gid)) == NULL)
		printf("gid=%ju, ", (uintmax_t)statp->st_gid);
	else
		printf("gid=\"%s\", ", grp->gr_name);
	printf("rdev=%ju, ", (uintmax_t)statp->st_rdev);
	printf("atime=");
	if (resolv == 0)
		printf("%jd", (intmax_t)statp->st_atim.tv_sec);
	else {
		tm = localtime(&statp->st_atim.tv_sec);
		(void)strftime(timestr, sizeof(timestr), TIME_FORMAT, tm);
		printf("\"%s\"", timestr);
	}
	if (statp->st_atim.tv_nsec != 0)
		printf(".%09ld, ", statp->st_atim.tv_nsec);
	else
		printf(", ");
	printf("stime=");
	if (resolv == 0)
		printf("%jd", (intmax_t)statp->st_mtim.tv_sec);
	else {
		tm = localtime(&statp->st_mtim.tv_sec);
		(void)strftime(timestr, sizeof(timestr), TIME_FORMAT, tm);
		printf("\"%s\"", timestr);
	}
	if (statp->st_mtim.tv_nsec != 0)
		printf(".%09ld, ", statp->st_mtim.tv_nsec);
	else
		printf(", ");
	printf("ctime=");
	if (resolv == 0)
		printf("%jd", (intmax_t)statp->st_ctim.tv_sec);
	else {
		tm = localtime(&statp->st_ctim.tv_sec);
		(void)strftime(timestr, sizeof(timestr), TIME_FORMAT, tm);
		printf("\"%s\"", timestr);
	}
	if (statp->st_ctim.tv_nsec != 0)
		printf(".%09ld, ", statp->st_ctim.tv_nsec);
	else
		printf(", ");
	printf("birthtime=");
	if (resolv == 0)
		printf("%jd", (intmax_t)statp->st_birthtim.tv_sec);
	else {
		tm = localtime(&statp->st_birthtim.tv_sec);
		(void)strftime(timestr, sizeof(timestr), TIME_FORMAT, tm);
		printf("\"%s\"", timestr);
	}
	if (statp->st_birthtim.tv_nsec != 0)
		printf(".%09ld, ", statp->st_birthtim.tv_nsec);
	else
		printf(", ");
	printf("size=%jd, blksize=%ju, blocks=%jd, flags=0x%x",
		(uintmax_t)statp->st_size, (uintmax_t)statp->st_blksize,
		(intmax_t)statp->st_blocks, statp->st_flags);
	printf(" }\n");
}

void
ktrstruct(char *buf, size_t buflen)
{
	char *name, *data;
	size_t namelen, datalen;
	int i;
	struct stat sb;
	struct sockaddr_storage ss;

	for (name = buf, namelen = 0;
	     namelen < buflen && name[namelen] != '\0';
	     ++namelen)
		/* nothing */;
	if (namelen == buflen)
		goto invalid;
	if (name[namelen] != '\0')
		goto invalid;
	data = buf + namelen + 1;
	datalen = buflen - namelen - 1;
	if (datalen == 0)
		goto invalid;
	/* sanity check */
	for (i = 0; i < namelen; ++i)
		if (!isalpha((unsigned char)name[i]))
			goto invalid;
	if (strcmp(name, "stat") == 0) {
		if (datalen != sizeof(struct stat))
			goto invalid;
		memcpy(&sb, data, datalen);
		ktrstat(&sb);
	} else if (strcmp(name, "sockaddr") == 0) {
		if (datalen > sizeof(ss))
			goto invalid;
		memcpy(&ss, data, datalen);
		if (datalen < sizeof(struct sockaddr) ||
		    datalen != ss.ss_len)
			goto invalid;
		ktrsockaddr((struct sockaddr *)&ss);
	} else {
		printf("unknown structure\n");
	}
	return;
invalid:
	printf("invalid record\n");
}

void
usage(void)
{

	fprintf(stderr, "usage: linux_kdump [-dEnlHRrsTA] [-f trfile] "
	    "[-m maxdata] [-p pid] [-t trstr]\n");
	exit(1);
}

void
sockfamilyname (int arg)
{
	if (arg == AF_UNSPEC) 
		printf("AF_UNSPEC");
	else if (arg == AF_LOCAL) 
		printf("AF_LOCAL");
	else if (arg == AF_UNIX) 
		printf("AF_UNIX");
	else if (arg == AF_INET) 
		printf("AF_INET");
	else if (arg == AF_IMPLINK) 
		printf("AF_IMPLINK");
	else if (arg == AF_PUP) 
		printf("AF_PUP");
	else if (arg == AF_CHAOS) 
		printf("AF_CHAOS");
	else if (arg == AF_NETBIOS) 
		printf("AF_NETBIOS");
	else if (arg == AF_ISO) 
		printf("AF_ISO");
	else if (arg == AF_OSI) 
		printf("AF_OSI");
	else if (arg == AF_ECMA) 
		printf("AF_ECMA");
	else if (arg == AF_DATAKIT) 
		printf("AF_DATAKIT");
	else if (arg == AF_CCITT) 
		printf("AF_CCITT");
	else if (arg == AF_SNA) 
		printf("AF_SNA");
	else if (arg == AF_DECnet) 
		printf("AF_DECnet");
	else if (arg == AF_DLI) 
		printf("AF_DLI");
	else if (arg == AF_LAT) 
		printf("AF_LAT");
	else if (arg == AF_HYLINK) 
		printf("AF_HYLINK");
	else if (arg == AF_APPLETALK) 
		printf("AF_APPLETALK");
	else if (arg == AF_ROUTE) 
		printf("AF_ROUTE");
	else if (arg == AF_LINK) 
		printf("AF_LINK");
	else if (arg == AF_COIP) 
		printf("AF_COIP");
	else if (arg == AF_CNT) 
		printf("AF_CNT");
	else if (arg == AF_IPX) 
		printf("AF_IPX");
	else if (arg == AF_SIP) 
		printf("AF_SIP");
	else if (arg == AF_ISDN) 
		printf("AF_ISDN");
	else if (arg == AF_E164) 
		printf("AF_E164");
	else if (arg == AF_INET6) 
		printf("AF_INET6");
	else if (arg == AF_NATM) 
		printf("AF_NATM");
	else if (arg == AF_ATM) 
		printf("AF_ATM");
	else if (arg == AF_NETGRAPH) 
		printf("AF_NETGRAPH");
	else if (arg == AF_SLOW) 
		printf("AF_SLOW");
	else if (arg == AF_SCLUSTER) 
		printf("AF_SCLUSTER");
	else if (arg == AF_ARP) 
		printf("AF_ARP");
	else if (arg == AF_BLUETOOTH) 
		printf("AF_BLUETOOTH");
	else if (arg == AF_IEEE80211) 
		printf("AF_IEEE80211");
	else if (arg == AF_MAX) 
		printf("AF_MAX");
	else if (arg == AF_VENDOR00) 
		printf("AF_VENDOR00");
	else if (arg == AF_VENDOR01) 
		printf("AF_VENDOR01");
	else if (arg == AF_VENDOR02) 
		printf("AF_VENDOR02");
	else if (arg == AF_VENDOR03) 
		printf("AF_VENDOR03");
	else if (arg == AF_VENDOR04) 
		printf("AF_VENDOR04");
	else if (arg == AF_VENDOR05) 
		printf("AF_VENDOR05");
	else if (arg == AF_VENDOR06) 
		printf("AF_VENDOR06");
	else if (arg == AF_VENDOR07) 
		printf("AF_VENDOR07");
	else if (arg == AF_VENDOR08) 
		printf("AF_VENDOR08");
	else if (arg == AF_VENDOR09) 
		printf("AF_VENDOR09");
	else if (arg == AF_VENDOR10) 
		printf("AF_VENDOR10");
	else if (arg == AF_VENDOR11) 
		printf("AF_VENDOR11");
	else if (arg == AF_VENDOR12) 
		printf("AF_VENDOR12");
	else if (arg == AF_VENDOR13) 
		printf("AF_VENDOR13");
	else if (arg == AF_VENDOR14) 
		printf("AF_VENDOR14");
	else if (arg == AF_VENDOR15) 
		printf("AF_VENDOR15");
	else if (arg == AF_VENDOR16) 
		printf("AF_VENDOR16");
	else if (arg == AF_VENDOR17) 
		printf("AF_VENDOR17");
	else if (arg == AF_VENDOR18) 
		printf("AF_VENDOR18");
	else if (arg == AF_VENDOR19) 
		printf("AF_VENDOR19");
	else if (arg == AF_VENDOR20) 
		printf("AF_VENDOR20");
	else if (arg == AF_VENDOR21) 
		printf("AF_VENDOR21");
	else if (arg == AF_VENDOR22) 
		printf("AF_VENDOR22");
	else if (arg == AF_VENDOR23) 
		printf("AF_VENDOR23");
	else if (arg == AF_VENDOR24) 
		printf("AF_VENDOR24");
	else if (arg == AF_VENDOR25) 
		printf("AF_VENDOR25");
	else if (arg == AF_VENDOR26) 
		printf("AF_VENDOR26");
	else if (arg == AF_VENDOR27) 
		printf("AF_VENDOR27");
	else if (arg == AF_VENDOR28) 
		printf("AF_VENDOR28");
	else if (arg == AF_VENDOR29) 
		printf("AF_VENDOR29");
	else if (arg == AF_VENDOR30) 
		printf("AF_VENDOR30");
	else if (arg == AF_VENDOR31) 
		printf("AF_VENDOR31");
	else if (arg == AF_VENDOR32) 
		printf("AF_VENDOR32");
	else if (arg == AF_VENDOR33) 
		printf("AF_VENDOR33");
	else if (arg == AF_VENDOR34) 
		printf("AF_VENDOR34");
	else if (arg == AF_VENDOR35) 
		printf("AF_VENDOR35");
	else if (arg == AF_VENDOR36) 
		printf("AF_VENDOR36");
	else if (arg == AF_VENDOR37) 
		printf("AF_VENDOR37");
	else if (arg == AF_VENDOR38) 
		printf("AF_VENDOR38");
	else if (arg == AF_VENDOR39) 
		printf("AF_VENDOR39");
	else if (arg == AF_VENDOR40) 
		printf("AF_VENDOR40");
	else if (arg == AF_VENDOR41) 
		printf("AF_VENDOR41");
	else if (arg == AF_VENDOR42) 
		printf("AF_VENDOR42");
	else if (arg == AF_VENDOR43) 
		printf("AF_VENDOR43");
	else if (arg == AF_VENDOR44) 
		printf("AF_VENDOR44");
	else if (arg == AF_VENDOR45) 
		printf("AF_VENDOR45");
	else if (arg == AF_VENDOR46) 
		printf("AF_VENDOR46");
	else if (arg == AF_VENDOR47) 
		printf("AF_VENDOR47");
	else /* Should not reach */
		(void)printf("<invalid=%ld>", (long)arg);
}
